package com.edgec.browserbackend.browser.task;

import com.alibaba.fastjson.JSONObject;
import com.edgec.browserbackend.account.domain.IpChargeRequestDto;
import com.edgec.browserbackend.account.domain.QueryIpUrlList;
import com.edgec.browserbackend.account.repository.QueryIpUrlListRepository;
import com.edgec.browserbackend.account.service.AccountService;
import com.edgec.browserbackend.browser.domain.IpResource;
import com.edgec.browserbackend.browser.domain.IpType;
import com.edgec.browserbackend.browser.dto.IpBuyResultDto;
import com.edgec.browserbackend.browser.dto.IpInfoResultDto;
import com.edgec.browserbackend.browser.dto.ShopRequestDto;
import com.edgec.browserbackend.browser.repository.IpResourceRepository;
import com.edgec.browserbackend.browser.service.IpAndShopService;
import com.edgec.browserbackend.common.commons.utils.NotifyUtils;
import com.edgec.browserbackend.common.utils.ThreadPoolUtils;
import com.edgec.browserbackend.common.utils.Trans;
import okhttp3.*;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;
import org.springframework.web.client.RestTemplate;

import java.io.IOException;
import java.time.Instant;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import static com.edgec.browserbackend.browser.service.Impl.IpResourceServiceImpl.genRandom;
import static com.edgec.browserbackend.browser.service.Impl.IpResourceServiceImpl.region;

@Component
public class BrowserTask {

    private static final Logger log = LoggerFactory.getLogger(BrowserTask.class);

    private static String CLOUDAMURL = "https://www.cloudam.cn";
    private static String TESTURL = "http://112.74.13.2";


    @Autowired
    private IpResourceRepository ipResourceRepository;

    @Autowired
    private AccountService accountService;

    @Autowired
    private IpAndShopService ipAndShopService;

    @Autowired
    private QueryIpUrlListRepository queryIpUrlListRepository;

    @Value("${spring.profiles.active}")
    private String profiles;

    public Map<String, String> buildGetHeader() {
        Map<String, String> headers = new HashMap<>();
        headers.put("Content-Type", "application/json");
        if (profiles.equals("dev") || profiles.equals("staging"))
            headers.put("Authorization", "Bearer oq5tg3gMsflHcK5iZ2741G5R30XYd9blyOqH9qeBItKtrzfTsGIoy8AsxqqNXdcm");
        else if (profiles.equals("prod"))
            headers.put("Authorization", "Bearer tKWsuHzcngf0RQPMss70f9jgymDIwgQ9zbLfESJdcou3pZSNWl7lNTzto8VQgwaO");
        return headers;
    }

    public HttpHeaders buildPostHeader() {
        HttpHeaders header = new HttpHeaders();
        header.setContentType(MediaType.APPLICATION_JSON);
        if (profiles.equals("dev") || profiles.equals("staging")) {
            header.setBearerAuth("oq5tg3gMsflHcK5iZ2741G5R30XYd9blyOqH9qeBItKtrzfTsGIoy8AsxqqNXdcm");
        } else if (profiles.equals("prod")) {
            header.setBearerAuth("tKWsuHzcngf0RQPMss70f9jgymDIwgQ9zbLfESJdcou3pZSNWl7lNTzto8VQgwaO");
        }
        return header;
    }

    private IpChargeRequestDto buildIpChargeRequestDto(IpResource request, int chargeType, int payMethod) {
        IpChargeRequestDto ipChargeRequestDto = new IpChargeRequestDto();
        ipChargeRequestDto.setAmount(1);
        ipChargeRequestDto.setChargeType(chargeType);
        ipChargeRequestDto.setRegion(request.getRegion());
        ipChargeRequestDto.setPeriod(request.getPeriod());
        ipChargeRequestDto.setUnit(request.getUnit());
        ipChargeRequestDto.setPayMethod(0);
        return ipChargeRequestDto;
    }


    @Scheduled(cron = "0 0/1 * * * ?")
    public void buyIpTasks() {
        String URL = (profiles.equals("dev") || profiles.equals("staging")) ? TESTURL : CLOUDAMURL;
        long time = Instant.now().minusSeconds(300).toEpochMilli();
        List<IpResource> ipResources = ipResourceRepository.sampleTasks(6, time);
        for (IpResource ipResource : ipResources) {
            long start = System.currentTimeMillis();
            CompletableFuture.runAsync(() -> {
                if (ipResourceRepository.lockTask(ipResource)) {
                    try {
                        boolean result = false;
                        RestTemplate restTemplate = new RestTemplate();
                        HttpHeaders header = buildPostHeader();
                        HashMap<String, Object> map = new HashMap<>();
                        map.put("name", ipResource.getUsername());
                        map.put("region", ipResource.getRegion());
                        map.put("period", String.valueOf(ipResource.getPeriod()));
                        map.put("provider", ipResource.getVendor());
                        map.put("unit", ipResource.getUnit());
                        map.put("amount", 1);
                        map.put("loginPassword", ipResource.getPassword());
                        map.put("startscript", "");
                        map.put("ipkeptperiod", 7);
                        map.put("persistSystemDiskOnTermination", false);
                        HttpEntity<Map<String, Object>> httpEntity = new HttpEntity<>(map, header);
                        IpBuyResultDto ipBuyResultDto = null;
                        try {
                            ipBuyResultDto = restTemplate.postForObject(URL + "/intelligroup/ipresources?accountId=browser", httpEntity, IpBuyResultDto.class);
                            if (StringUtils.isNotBlank(ipBuyResultDto.getErrorCode())) {
                                log.error("fail to buy ip");
                                log.error(ipBuyResultDto.getErrorCode());
                            }
                            if (ipBuyResultDto != null && ipBuyResultDto.getIplist() != null && ipBuyResultDto.getIplist().size() >= 1) {
                                AtomicInteger index = new AtomicInteger();
                                ipBuyResultDto.getIplist().forEach(x -> {
                                    if (ipResource != null) {
                                        ipResource.setAddr(x.getIp());
                                        ipResource.setStatus(3);
                                        ipResource.setValidTime(Instant.parse(x.getValidTill()).toEpochMilli());
                                        if (StringUtils.isNotBlank(ipResource.getRegion()) && region.contains(ipResource.getRegion())) {
                                            ipResource.setProxyUsername(ipResource.getAddr());
                                            ipResource.setProxyPassword(genRandom(3, 12));
                                            ipResource.setSpecialLine(true);
                                        }
                                        ipResourceRepository.save(ipResource);
                                    } else {
                                        log.error("no ipResource");
                                    }
                                    index.getAndIncrement();
                                });
                                result = true;
                            }
                        } catch (Exception e) {
                            log.error(e.getMessage());
                            result = false;
                        }
                        if (result == false && (ipResource.getPurchasedTime() < Instant.now().minusSeconds(7200).toEpochMilli())) {
                            IpChargeRequestDto ipChargeRequestDto = buildIpChargeRequestDto(ipResource, 3, 0);
                            accountService.chargeByMoney(ipResource.getUsername(), -ipResource.getPrice(), ipChargeRequestDto);
                            if (ipResource.getShopId() != null) {
                                ShopRequestDto shopRequestDto = new ShopRequestDto();
                                shopRequestDto.setIpId(ipResource.getId());
                                shopRequestDto.setShopId(ipResource.getShopId());
                                ipAndShopService.unBindShop(ipResource.getUsername(), shopRequestDto);
                            }
                            ipResourceRepository.deleteById(ipResource.getId());
                        }
                    } finally {
                        long end = System.currentTimeMillis();
                        log.debug("buyIpTask {} execution time is: " + (end - start) / 1000 + "s", ipResource.getId());
                        try {
                            ipResourceRepository.unLockTask(ipResource.getId());
                        } catch (Throwable th) {
                            log.error("unlock failed", th);
                            //try again
                            ipResourceRepository.unLockTask(ipResource.getId());
                        }
                    }
                }
            }, ThreadPoolUtils.buyIpTasksPool);
        }
    }

    @Scheduled(cron = "0 0/1 * * * ?")
    public void queryIpTasks() {
        String URL = (profiles.equals("dev") || profiles.equals("staging")) ? TESTURL : CLOUDAMURL;
        long time = Instant.now().minusSeconds(300).toEpochMilli();
        List<IpResource> ipResources = ipResourceRepository.sampleTasks(3, time);
        for (IpResource ipResource : ipResources) {
            long start = System.currentTimeMillis();
            CompletableFuture.runAsync(() -> {
                if (ipResourceRepository.lockTask(ipResource)) {
                    try {
                        String url = URL + "/ecc/ipinfo?accountId=browser&ip=" + ipResource.getAddr();
                        Map<String, String> header = buildGetHeader();
                        String rs = HttpClientutils.doGet(url, header);
                        IpInfoResultDto ipInfoResultDto = JSONObject.parseObject(rs, IpInfoResultDto.class);
                        if (ipInfoResultDto != null && StringUtils.isBlank(ipInfoResultDto.getErrorCode())) {
                            if (StringUtils.isNotEmpty(ipInfoResultDto.getStatus())) {
                                ipResource.setStatus(0);
                                ipResourceRepository.save(ipResource);
                            }
                        }
                    } catch (Exception e) {
                        log.error(e.getMessage(), e);
                        NotifyUtils.sendMessage("浏览器后端 queryIpTasks() 又炸了，赶紧看啊", e, NotifyUtils.MsgType.WEBHOOK);
                    } finally {
                        long end = System.currentTimeMillis();
                        log.debug("queryIpTask {} execution time is: " + (end - start) / 1000 + "s", ipResource.getId());
                        try {
                            ipResourceRepository.unLockTask(ipResource.getId());
                        } catch (Throwable th) {
                            log.error("unlock failed", th);
                            //try again
                            ipResourceRepository.unLockTask(ipResource.getId());
                        }
                    }
                }
            }, ThreadPoolUtils.queryIpTasksPool);
        }
    }

    @Scheduled(cron = "0 0/5 * * * ?")
    public void healthCheck() {
        List<IpResource> ipResources = ipResourceRepository.sampleTasks(Arrays.asList(0, 2));
        for (IpResource ipResource : ipResources) {
            long start = System.currentTimeMillis();
            if (ipResourceRepository.healthLock(ipResource)) {
                try {
                    QueryIpUrlList queryIpUrlList = queryIpUrlListRepository.findAll().get(0);
                    if (ipResource.isSpecialLine()) {
                        Trans trans = new Trans(ipResource.getProxyUsername(), ipResource.getProxyPassword());
                        String sp_result = trans.get(queryIpUrlList.getUrl());
                        int failTime = 0;
                        while (!sp_result.contains(ipResource.getAddr())) {
                            if (failTime > 5) {
                                NotifyUtils.sendMessage("防关联浏览器 ip " + ipResource.getAddr() + " 专线代理异常", NotifyUtils.MsgType.WEBHOOK);
                                log.error("防关联浏览器 ip " + ipResource.getAddr() + " 专线代理异常 " + sp_result);
                                break;
                            }
                            failTime ++;
                            Thread.sleep(2000);
                            sp_result = trans.get(queryIpUrlList.getUrl());
                        }
                    } else if (ipResource.getIpType() == IpType.VENDOR){
                        String result;
                        Trans trans = new Trans(ipResource.getAddr(), Integer.valueOf(ipResource.getPort().size() > 1?ipResource.getPort().get(1):ipResource.getPort().get(0)), ipResource.getUsername(), ipResource.getPassword());
                        result = trans.get(queryIpUrlList.getUrl());
                        if (!result.contains(ipResource.getAddr())) {
                            int failTime = 0;
                            while (!result.contains(ipResource.getAddr())) {
                                if (failTime > 5) {
                                    NotifyUtils.sendMessage("防关联浏览器 ip " + ipResource.getAddr() + " 代理异常", NotifyUtils.MsgType.WEBHOOK);
                                    log.error("防关联浏览器 ip " + ipResource.getAddr() + " 代理异常 " + result);
                                    break;
                                }
                                failTime ++;
                                Thread.sleep(2000);
                                result = trans.get(queryIpUrlList.getUrl());
                            }
                        }
                    }
                } catch (Exception e) {
                    NotifyUtils.sendMessage("防关联浏览器 ip " + ipResource.getAddr() + " 代理异常", e, NotifyUtils.MsgType.WEBHOOK);
                    log.error(e.getMessage(), e);
                } finally {
                    long end = System.currentTimeMillis();
                    log.debug("queryIpTask {} execution time is: " + (end - start) / 1000 + "s", ipResource.getId());
                    try {
                        ipResourceRepository.unLockHealth(ipResource.getId());
                    } catch (Throwable th) {
                        log.error("unlock failed", th);
                        //try again
                        ipResourceRepository.unLockHealth(ipResource.getId());
                    }
                }
            }
        }
    }


    public static class HttpClientutils {
        static OkHttpClient.Builder builder = new OkHttpClient.Builder();

        static {
            builder.connectTimeout(10, TimeUnit.SECONDS);
            builder.readTimeout(10, TimeUnit.SECONDS);
        }

        static OkHttpClient client = new OkHttpClient(builder);

        public static String doGet(String url, Map<String, String> headers) throws IOException {
            Headers httpHeaders = Headers.of(headers);
            Request request = new Request.Builder()
                    .get()
                    .url(url)
                    .headers(httpHeaders)
                    .build();
            Call call = client.newCall(request);
            Response response = call.execute();
            ResponseBody responseBody = response.body();
            return responseBody.string();
        }

    }


    public static void main(String[] args) throws IOException {
        String url = CLOUDAMURL + "/ecc/ipinfo?accountId=browser&ip=" + "100.25.101.201";
        String profiles = "prod";
        Map<String, String> headers = new HashMap<>();
        headers.put("Content-Type", "application/json");
        if (profiles.equals("dev") || profiles.equals("staging"))
            headers.put("Authorization", "Bearer oq5tg3gMsflHcK5iZ2741G5R30XYd9blyOqH9qeBItKtrzfTsGIoy8AsxqqNXdcm");
        else if (profiles.equals("prod"))
            headers.put("Authorization", "Bearer tKWsuHzcngf0RQPMss70f9jgymDIwgQ9zbLfESJdcou3pZSNWl7lNTzto8VQgwaO");

        String rs = HttpClientutils.doGet(url, headers);
        System.out.println(rs);
        IpInfoResultDto ipInfoResultDto = JSONObject.parseObject(rs, IpInfoResultDto.class);
        System.out.println(ipInfoResultDto);
    }


}
